/* Copyright (c) 2005 Dirk Jagdmann <doj@cubic.org>

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you
       must not claim that you wrote the original software. If you use
       this software in a product, an acknowledgment in the product
       documentation would be appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and
       must not be misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
       distribution. */

#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <stdint.h>
#include <stdio.h>
#include "memorymap.hpp"
#include "memorymap.cpp"
#include <iostream>
#include <string>
#include <glib.h>

#pragma pack (push, 1)
struct T64head
{
  char id[32];
  uint16_t version;
  uint16_t maxentries;
  uint16_t entries;
  uint16_t unused;
  char name[24];
};

struct T64entry
{
  uint8_t filetype_c64;
  uint8_t filetype_1541;
  uint16_t startaddr;
  uint16_t endaddr;
  uint16_t unused1;
  uint32_t offset;
  uint32_t unused2;
  char filename[16];
};
#pragma pack (pop)

T64entry nullentry;		// a zero filled entry

char PETchar(const int c)
{
  const char petchar_[] = {
    /* 00 */ '@','a','b','c','d','e','f','g',
    /* 08 */ 'h','i','j','k','l','m','n','o',
    /* 10 */ 'p','q','r','s','t','u','v','w',
    /* 18 */ 'x','y','z','[','',']','^','.',
    /* 20 */ ' ','!','"','#','$','%','&','\'',
    /* 28 */ '(',')','*','+',',','-','.','/',
    /* 30 */ '0','1','2','3','4','5','6','7',
    /* 38 */ '8','9',':',';','<','=','>','?',
    /* 40 */ '.','A','B','C','D','E','F','G',
    /* 48 */ 'H','I','J','K','L','M','N','O',
    /* 50 */ 'P','Q','R','S','T','U','V','W',
    /* 58 */ 'X','Y','Z',/*'.','.','.','.','.',*/
#if 0
    /* 60 */ ' ','.','.','.','.','.','.','.',
    /* 68 */ '.','.','.','.','.','.','.','.',
    /* 70 */ '.','.','.','.','.','.','.','.',
    /* 78 */ '.','.','.','.','.','.','.','.',
#endif
  };
  return (c>=0 && c<sizeof(petchar_))?petchar_[c]:'.';
}

// remove unwanted characters from s
std::string fixstring(const std::string& s)
{
  std::string d;
  for(std::string::const_iterator i=s.begin(); i!=s.end(); ++i)
    {
      if(*i == '\'')
	d+="'\"'\"'";
      else if(isprint(*i))
	d+=*i;
    }
  return d;
}

int main(int argc, char **argv)
{
  // check command line
  if(argc != 2)
    {
      std::cerr << "usage: t64dump <filename>\n"
		<< "http://llg.cubic.org/tools#t64dump\n"
		<< "version 2005-07-06\n";
      return 1;
    }

  // load t64
  doj::memorymap_ptr<T64head> p(argv[1]);
  if(!p)
    {
      std::cerr << "could not open " << argv[1] << ":" << strerror(errno) << std::endl;
      return 1;
    }

  T64head h; memcpy(&h, p.get(), sizeof(T64head));

  // check header of t64
  if(memcmp(h.id, "C64S tape file\r\n", 16)
     && memcmp(h.id, "C64 tape image file\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0", 32)
     && memcmp(h.id, "C64S tape image file\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0\x0", 32) )
    {
      std::cerr << argv[1] << " is not a T64\n";
      return 1;
    }

  h.version=GINT16_FROM_LE(h.version);
  if(!(h.version == 0x100 || h.version == 0x101))
    {
      std::cerr << argv[1] << " unknown version\n";
      return 1;
    }

  h.entries=GINT16_FROM_LE(h.entries);
  h.maxentries=GINT16_FROM_LE(h.maxentries);
  if(h.entries > h.maxentries)
    {
      std::cerr << argv[1] << " entries(" << h.entries << ") exceed maxentries(" << h.maxentries << ")\n";
      return 1;
    }

  const int entries=(h.entries==0) ? h.maxentries : h.entries;

  // construct d64 filename
  std::string image=fixstring(argv[1]);
  const int s=image.size();
  if(s>3 && image[s-1]=='4' && image[s-2]=='6')
    image[s-3]='d';
  else
    image+=".d64";

  // construct d64 imagename
  char nm_[25]; memcpy(nm_, h.name, 24); nm_[24]=0;
  for(int i=23; i>=0; --i)
    if(nm_[i]!=' ')
      break;
    else
      nm_[i]=0;
  const std::string nm=fixstring(nm_);
  std::cout << "#!/bin/sh\n"
	    << "# " << nm << std::endl
	    << "# entries " << entries << std::endl
	    << "c1541 -format '" << nm << ",aa' d64 '" << image << "'\n";

  for(int i=0; i!=entries; ++i)
    {
      const T64entry *e=reinterpret_cast<const T64entry*>(p.get()+sizeof(T64head)+i*sizeof(T64entry));

      // if entry is completly zero we ignore it
      if(!memcmp(e, &nullentry, sizeof(T64entry)))
	continue;

      // construct filename
      char fn_[17]; fn_[16]=0;
      for(int j=15, flag=1; j>=0; --j)
	if(flag && e->filename[j]==0x20)
	  fn_[j]=0;
	else
	  {
	    fn_[j]=PETchar(e->filename[j]);
	    flag=0;
	  }
      const std::string fn=fixstring(fn_);

      std::cout << "# " << i << "\t" << fn << "\t";

      // extract startaddr and length of file
      uint16_t startaddr=GINT16_FROM_LE(e->startaddr);
      uint16_t endaddr=GINT16_FROM_LE(e->endaddr);
      if(!(startaddr < endaddr || (startaddr==0 && endaddr==0)))
	{
	  std::cout << "illegal startaddr/endaddr\n";
	  continue;
	}

      int length=(startaddr==0 && endaddr==0) ? 0x10000 : (endaddr - startaddr);
      std::cout << startaddr << "\t" << length << "\t";

      // determine filetype
      if(e->filetype_1541==0 && e->filetype_c64>1)
	{
	  std::cout << "FRZ";
	}
      else
	{
	  if(e->filetype_1541 & 0x30) std::cout << ">";
	  switch(e->filetype_1541 & 7)
	    {
	    case 0: std::cout << "DEL"; break;
	    case 1: std::cout << "SEQ"; break;
	    case 2: std::cout << "PRG"; break;
	    case 3: std::cout << "USR"; break;
	    case 4: std::cout << "REL"; break;
	    default:std::cout << "ILL"; break;
	    }
	  if(e->filetype_1541 & 0x10) std::cout << "?";
	  if(e->filetype_1541 & 0x20) std::cout << "@";
	  if(e->filetype_1541 & 0x40) std::cout << "*";
	}

      std::cout << std::endl;

      // save file
      FILE *f=fopen(fn.c_str(), "wb");
      if(f)
	{
	  fwrite(&startaddr, 1, 2, f);
	  fwrite(p.get()+GINT32_FROM_LE(e->offset), 1, length, f);
	  fclose(f);
	  std::cout << "c1541 '" << image << "' -write '" << fn << "'\n"
		    << "rm '" << fn << "'\n";
	}
      else
	std::cerr << fn << ":" << strerror(errno) << std::endl;
    }

  return 0;
}
